# This example is based on the article Multi-party Secure Position
# Determination by Tord Ingolf Reistad in 2006. In his example three
# three anchor nodes A, B, C and one floater node P want to securely
# compute the geographical position of P. The privacy of location
# data is highly valued
#
# Refer to the paper for an explanation of the protocol:
# http://www.nik.no/2006/Reistad.pdf
#
# Give a player configuration file as a command line argument or run
# the example with '--help' for help with the command line options.

from optparse import OptionParser
from twisted.internet import reactor

from viff.field import GF
from viff.runtime import Runtime, create_runtime, gather_shares
from viff.comparison import Toft05Runtime
from viff.config import load_config
from viff.util import rand, find_prime
from math import sqrt

class Protocol:
    """We start by defining the protocol, it will be started
    at the bottom of the file."""

    def __init__(self, runtime):
        # Save the Runtime for later use
        self.runtime = runtime

        # This is the values we will use in the protocol.
        self.x = rand.randint(1, 200)
        self.y = rand.randint(1, 200)
        # The variable ra1 is made to obfuscate the result for some
        # of the parties
        self.ra1 = rand.randint(1, 2009999999)
        # For convenience we decide that the coordinate will be (100,100)
        self.distance = sqrt((self.x - 100)**2 + (self.y -100)**2) + 0.5
        self.distance2 = int(self.distance)

        print "x:", self.x, "y:", self.y
        print "distance:", self.distance2
        print "ra1:", self.ra1

        # For the comparison protocol to work, we need a field modulus
        # bigger than 2**(l+1) + 2**(l+k+1), where the bit length of
        # the input numbers is l and k is the security parameter.
        # Further more, the prime must be a Blum prime (a prime p such
        # that p % 4 == 3 holds). The find_prime function lets us find
        # a suitable prime.
        l = runtime.options.bit_length
        k = runtime.options.security_parameter
        Zp = GF(find_prime(2**(l + 1) + 2**(l + k + 1), blum=True))
        test = find_prime(2**(l + 1) + 2**(l + k + 1), blum=True)
        print "prime:", test

        # We must secret share our input with the other parties.
        xa, xb, xc = runtime.shamir_share([1, 2, 3], Zp, self.x)
        ya, yb, yc = runtime.shamir_share([1, 2, 3], Zp, self.y)
        ap, bp, cp = runtime.shamir_share([1, 2, 3], Zp, self.distance2)

        # Here it is only player 1 that shares his ra1 variable
        if runtime.id == 1:
            r1 = runtime.shamir_share([1], Zp, self.ra1)
        else:
            r1 = runtime.shamir_share([1], Zp)

        # Now that everybody has secret shared their inputs we can
        # do some calculations
        print "Trying to calculate a1, b1, etc"  
        a1 = 2 * (xa - xb)
        b1 = 2 * (ya - yb)
        c1 = xa*xa - xb*xb + ya*ya - yb*yb + bp*bp - ap*ap
        a2 = 2 * (xa - xc)
        b2 = 2 * (ya - yc)
        c2 = xa*xa - xc*xc + ya*ya - yc*yc + cp*cp - ap*ap

        print "Trying to calculate xp_up, xp_do, etc"        
        xp_up = c1*b2 - c2*b1 + r1
        xp_do = a1*b2 - b1*a2
        yp_up = c1*a2 - c2*a1
        yp_do = a1*b2 - b1*a2

        # The results are secret shared, so we must open them before
        # we can do anything usefull with them.
        open_xp_up = runtime.open(xp_up)
        open_xp_do = runtime.open(xp_do)
        open_yp_up = runtime.open(yp_up)
        open_yp_do = runtime.open(yp_do)

        # We will now gather the results and call the
        # self.results_ready method when they have all been received.
        results = gather_shares([open_xp_up, open_xp_do, open_yp_up, open_yp_do])
        results.addCallback(self.results_ready)

        # We can add more callbacks to the callback chain in results.
        # These are called in sequence when self.results_ready is
        # finished. The first callback acts like a barrier and makes
        # all players wait on each other.
        #
        # The callbacks are always called with an argument equal to
        # the return value of the preceeding callback. We do not need
        # the argument (which is None since self.results_ready does
        # not return anything), so we throw it away using a lambda
        # expressions which ignores its first argument.
        results.addCallback(lambda _: runtime.synchronize())
        # The next callback shuts the runtime down, killing the
        # connections between the players.
        results.addCallback(lambda _: runtime.shutdown())      

    def results_ready(self, results):
        # Since this method is called as a callback above, the results
        # variable will contain actual field elements, not just
        # Shares. That makes it very easy to work on them.

        # Let us start by unpacking the list.
        open_xp_up = results[0] - self.ra1
        open_xp_do = results[1]
        open_yp_up = results[2]
        open_yp_do = results[3]

        def convert(x):
            #Convert takes a field element and converts it to an integer
            if (x.value < x.modulus/2):
                print x.value
                return x.value
            else:
                print x.value - x.modulus
                return x.value - x.modulus

        xp_up = convert(open_xp_up)
        xp_do = convert(open_xp_do)
        yp_up = convert(open_yp_up)
        yp_do = convert(open_yp_do)
        
        print "xp_up:", xp_up, "xp_do:", xp_do, "yp_up:", yp_up, "yp_do:", yp_do

        xp = xp_up / xp_do
        yp = -yp_up / yp_do
        print "xp:", xp, "yp:", yp

# Parse command line arguments.
parser = OptionParser()
Runtime.add_options(parser)
options, args = parser.parse_args()

if len(args) == 0:
    parser.error("you must specify a config file")
else:
    id, players = load_config(args[0])

# Create a deferred Runtime and ask it to run our protocol when ready.
pre_runtime = create_runtime(id, players, 1, options, Toft05Runtime)
pre_runtime.addCallback(Protocol)

# Start the Twisted event loop.
reactor.run()